//-< PessimisticMetaobject.java >------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      1-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update:  9-Oct-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Metaobject for pessimistic scheme of synchronization
//-------------------------------------------------------------------*--------*
namespace Goods
{
    using System;
	
    /// <summary>  <code>PessimisticMetaobject</code>  for pessimistic scheme of synchronization.
    /// Or would it be realistic. 
    /// Or Multiple Reader, one writer.
    /// 
    /// </summary>
    /// <author> K.A. Knizhnik
    /// </author>
    /// <version> 1.0
    /// 
    /// </version>
    public class PessimisticMetaobject:BasicMetaobject
    {
		
        /// <summary>  <code>beginWriteAccess</code> locks the object. That is, the objects storage 
        /// manager is requested to objeain the lock from the server.
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> to be modified.
        /// 
        /// </param>
        protected internal override void  beginWriteAccess(ObjectHeader obj)
        {
            if ((obj.state & ObjectState.XLOCKED) == 0)
            {
                obj.storage.lockObject(obj, Protocol.lck_exclusive, Protocol.lckattr_wait);
                obj.state |= ObjectState.XLOCKED;
            }
        }
		
        /// <summary>  <code>beginReadAccess</code> is a no-op. So, no locking for read's. Multiple
        /// readers, one writer.
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> object to be read.
        /// 
        /// </param>
        protected internal override void  beginReadAccess(ObjectHeader obj)
        {
        }
		
        /// <summary>  <code>endAccess</code> adds the object to the cahce managers transaction list
        ///
        /// </summary>
        /// <param name="obj">a <code>ObjectHeader</code> value
        /// 
        /// </param>
        protected internal override void  endAccess(ObjectHeader obj)
        {
            ObjectState state = obj.state;
            if ((state & ObjectState.TRANSWRITE) == 0 && (state & (ObjectState.DIRTY | ObjectState.XLOCKED)) != 0)
            {
                CacheManager mng = CacheManager.GetCacheManager();
                mng.addToTransaction(obj);
            }
        }
    }
}