//-< Protocol.java >-------------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      1-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update: 10-Oct-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Description of client-server communication protocol
//-------------------------------------------------------------------*--------*
namespace Goods
{
    using System;
	
    class Protocol
    {
        internal const int RAW_CPID = 0x00001;
        internal const int ROOT_OPID = 0x10000;
		
        internal const int ALLOC_ALIGNED = (int) (- (0x100000000 - 0x80000000));
		
        //
        // Object reference components
        //
        internal const int ref_sid_offs = 0;
        internal const int ref_opid_offs = 2;
		
        internal const int ref_size = 6;
		
        //
        // Offsets of object header fields
        //
        internal const int hdr_flags_offs = 0;
        internal const int hdr_opid_offs = 4;
        internal const int hdr_cpid_offs = 8;
        internal const int hdr_sid_offs = 10;
        internal const int hdr_size_offs = 12;
		
        internal const int hdr_size = 16;
		
        //
        // Types of object fields
        //
        internal const int fld_structure = 0;
        internal const int fld_reference = 1;
        internal const int fld_signed_integer = 2;
        internal const int fld_unsigned_integer = 3;
        internal const int fld_real = 4;
        internal const int fld_string = 5;
        internal const int fld_raw_binary = 6;
		
        //
        // Field descriptor fields offsets
        //
        internal const int fds_type_offs = 0;
        internal const int fds_name_offs = 2;
        internal const int fds_size_offs = 4;
        internal const int fds_nitems_offs = 8;
        internal const int fds_next_offs = 12;
		
        internal const int fds_size = 16;
		
        //
        // Class descriptor fields offsets
        //
        internal const int cds_fixed_size_offs = 0;
        internal const int cds_varying_size_offs = 4;
        internal const int cds_n_fixed_refs_offs = 8;
        internal const int cds_n_varying_refs_offs = 12;
        internal const int cds_n_fields_offs = 16;
        internal const int cds_total_names_size_offs = 20;
		
        internal const int cds_header_size = 24;
		
        //
        // Lock request types
        //
        internal const int lck_none = 0; // No lock
        internal const int lck_shared = 1; // Prevent object from been modified or
        // exclusivly locked by other processes
        internal const int lck_exclusive = 2; // Prevent object from been modified or
        // locked (shared or exlusive) by other processes
		
        //
        // Lock attributes
        //
        internal const int lckattr_wait = 0;
        internal const int lckattr_nowait = 1; // Lock request will not block, 
        // if granting lock is impossible, request will be immediately refused 
		
        //
        // Transaction object flags 
        //
        internal const int tof_none = 0;
        internal const int tof_update = 1; // request to update object
        internal const int tof_validate = 2; // request to validate transaction object at server
        internal const int tof_unlock = 4; // request to unlock transaction object at the end of transaction
        internal const int tof_unlock_exl = 8; // request to release exclusive lock at the end of transaction
        //
        // load object flags
        //
        internal const int lof_none = 0;
        internal const int lof_copy = 1; // request another copy of object
        internal const int lof_cluster = 2; // enable sending cluster of objects
		
		
        // 
        // Command request codes
        // 
        internal const byte cmd_load = 0; // client request object from server
        internal const byte cmd_object = 1; // server sent object to client
        internal const byte cmd_forget = 2; // client remove reference to object
        internal const byte cmd_throw = 3; // client throw away instance of object from the cache
        internal const byte cmd_invalidate = 4; // server inform client about object modification
		
        internal const byte cmd_getclass = 5; // get class descriptor from server
        internal const byte cmd_classdesc = 6; // server send class descriptor to client
		
        internal const byte cmd_putclass = 7; // register new class at server 
        internal const byte cmd_classid = 8; // class identifier returned to client 
		
        internal const byte cmd_modclass = 9; // modify existed class at server
		
        internal const byte cmd_lock = 10; // request to server to lock object
        internal const byte cmd_lockresult = 11; // result of applying lock request 
        internal const byte cmd_unlock = 12; // request to server to unlock object
		
        internal const byte cmd_transaction = 13; // send transaction to coordinator
        internal const byte cmd_subtransact = 14; // send local part of global trans.
        internal const byte cmd_transresult = 15; // coordinator returns trans. status
        internal const byte cmd_tmsync = 16; // byteerserver transaction sync. message
		
        internal const byte cmd_login = 17; // client login at server 
        internal const byte cmd_logout = 18; // client finish the session
        internal const byte cmd_connect = 19; // one server connects to other 
        internal const byte cmd_bye = 20; // server notifies client  about it's termination
        internal const byte cmd_ok = 21; // successful authorization 
        internal const byte cmd_refused = 22; // authorization procedure is failed at server
        internal const byte cmd_alloc = 23; // request server to allocate object
        internal const byte cmd_location = 24; // server returns address of allocated object
        internal const byte cmd_free = 25; // client request to free object
        internal const byte cmd_bulk_alloc = 26; // client request server to allocate several objects and reserve OIDs 
        internal const byte cmd_opids = 27; // object identifiers reserved by server for the client
        internal const byte cmd_gcsync = 28; // interserver garbage collection synchonization
		
        internal const byte cmd_get_size = 29; // get storage size
		
        //
        // Request fields offsets
        //
        internal const int req_cmd_offs = 0;
        internal const int req_arg1_offs = 1;
        internal const int req_arg2_offs = 2;
        internal const int req_arg3_offs = 4;
        internal const int req_arg4_offs = 8;
        internal const int req_arg5_offs = 12;
		
        internal const int req_type_offs = 1;
        internal const int req_n_servers_offs = 1;
        internal const int req_align_offs = 1;
        internal const int req_status_offs = 1;
        internal const int req_coordinator_offs = 2;
        internal const int req_flags_offs = 2;
        internal const int req_cpid_offs = 2;
        internal const int req_sid_offs = 2;
        internal const int req_attr_offs = 2;
        internal const int req_size_offs = 4;
        internal const int req_opid_offs = 4;
        internal const int req_extra_offs = 8;
        internal const int req_reserve_offs = 8;
        internal const int req_tid_offs = 8;
        internal const int req_crc_offs = 12;
		
        internal const int req_size = 16;
		
        internal static int unpack2(byte[] data, int offs)
        {
            return ((sbyte)data[offs] << 8) + data[offs + 1];
        }
        internal static int unpack4(byte[] data, int offs)
        {
            return ((sbyte)data[offs] << 24) + (data[offs + 1] << 16) + (data[offs + 2] << 8) + data[offs + 3];
        }
        internal static long unpack8(byte[] data, int offs)
        {
            return ((long)unpack4(data, offs) << 32) + (uint)unpack4(data, offs + 4);
        }
        internal static String unpackAscii(byte[] data, int offs)
        {
            int eos = offs;
            while (data[eos] != 0)
            {
                eos += 1;
            }
            char[] chars = new char[eos-offs];
            for (int i = 0; i < chars.Length; i++) 
            { 
                chars[i] = (char)data[offs+i];
            }
            return new String(chars);
        }
		
        internal static int pack2(byte[] data, int offs, int val)
        {
            data[offs++] = (byte) (val >> 8);
            data[offs++] = (byte) val;
            return offs;
        }
        internal static int pack4(byte[] data, int offs, int val)
        {
            data[offs++] = (byte) (val >> 24);
            data[offs++] = (byte) (val >> 16);
            data[offs++] = (byte) (val >> 8);
            data[offs++] = (byte) val;
            return offs;
        }
        internal static int pack8(byte[] data, int offs, long val)
        {
            return pack4(data, pack4(data, offs, (int) (val >> 32)), (int) val);
        }
		
        internal static int packAscii(byte[] data, int offs, String s)
        {
            int length = s.Length;
            int i;
            int j;
            i = 0;
            j = offs;
            while (i < length)
            {
                data[j] = (byte) s[i];
                i++;
                j++;
            }
            data[offs + length] = 0;
            return offs + length + 1;
        }


        internal static int packDate(byte[] data, int offs, DateTime dt) 
        { 
            return pack8(data, offs, dt.Ticks == -1 ? dt.Ticks : (dt.Ticks - 621355968000000000L) / 10000);
        }

        internal static DateTime NewDateTime(long milliseconds) 
        {
            return new DateTime(milliseconds == -1 ? milliseconds : milliseconds*10000 + 621355968000000000L);
        }         

        static uint[] crc32table = {
            0x00000000, 0x04C11DB7, 0x09823B6E, 0x0D4326D9,
            0x130476DC, 0x17C56B6B, 0x1A864DB2, 0x1E475005,
            0x2608EDB8, 0x22C9F00F, 0x2F8AD6D6, 0x2B4BCB61,
            0x350C9B64, 0x31CD86D3, 0x3C8EA00A, 0x384FBDBD,
            0x4C11DB70, 0x48D0C6C7, 0x4593E01E, 0x4152FDA9,
            0x5F15ADAC, 0x5BD4B01B, 0x569796C2, 0x52568B75,
            0x6A1936C8, 0x6ED82B7F, 0x639B0DA6, 0x675A1011,
            0x791D4014, 0x7DDC5DA3, 0x709F7B7A, 0x745E66CD,
            0x9823B6E0, 0x9CE2AB57, 0x91A18D8E, 0x95609039,
            0x8B27C03C, 0x8FE6DD8B, 0x82A5FB52, 0x8664E6E5,
            0xBE2B5B58, 0xBAEA46EF, 0xB7A96036, 0xB3687D81,
            0xAD2F2D84, 0xA9EE3033, 0xA4AD16EA, 0xA06C0B5D,
            0xD4326D90, 0xD0F37027, 0xDDB056FE, 0xD9714B49,
            0xC7361B4C, 0xC3F706FB, 0xCEB42022, 0xCA753D95,
            0xF23A8028, 0xF6FB9D9F, 0xFBB8BB46, 0xFF79A6F1,
            0xE13EF6F4, 0xE5FFEB43, 0xE8BCCD9A, 0xEC7DD02D,
            0x34867077, 0x30476DC0, 0x3D044B19, 0x39C556AE,
            0x278206AB, 0x23431B1C, 0x2E003DC5, 0x2AC12072,
            0x128E9DCF, 0x164F8078, 0x1B0CA6A1, 0x1FCDBB16,
            0x018AEB13, 0x054BF6A4, 0x0808D07D, 0x0CC9CDCA,
            0x7897AB07, 0x7C56B6B0, 0x71159069, 0x75D48DDE,
            0x6B93DDDB, 0x6F52C06C, 0x6211E6B5, 0x66D0FB02,
            0x5E9F46BF, 0x5A5E5B08, 0x571D7DD1, 0x53DC6066,
            0x4D9B3063, 0x495A2DD4, 0x44190B0D, 0x40D816BA,
            0xACA5C697, 0xA864DB20, 0xA527FDF9, 0xA1E6E04E,
            0xBFA1B04B, 0xBB60ADFC, 0xB6238B25, 0xB2E29692,
            0x8AAD2B2F, 0x8E6C3698, 0x832F1041, 0x87EE0DF6,
            0x99A95DF3, 0x9D684044, 0x902B669D, 0x94EA7B2A,
            0xE0B41DE7, 0xE4750050, 0xE9362689, 0xEDF73B3E,
            0xF3B06B3B, 0xF771768C, 0xFA325055, 0xFEF34DE2,
            0xC6BCF05F, 0xC27DEDE8, 0xCF3ECB31, 0xCBFFD686,
            0xD5B88683, 0xD1799B34, 0xDC3ABDED, 0xD8FBA05A,
            0x690CE0EE, 0x6DCDFD59, 0x608EDB80, 0x644FC637,
            0x7A089632, 0x7EC98B85, 0x738AAD5C, 0x774BB0EB,
            0x4F040D56, 0x4BC510E1, 0x46863638, 0x42472B8F,
            0x5C007B8A, 0x58C1663D, 0x558240E4, 0x51435D53,
            0x251D3B9E, 0x21DC2629, 0x2C9F00F0, 0x285E1D47,
            0x36194D42, 0x32D850F5, 0x3F9B762C, 0x3B5A6B9B,
            0x0315D626, 0x07D4CB91, 0x0A97ED48, 0x0E56F0FF,
            0x1011A0FA, 0x14D0BD4D, 0x19939B94, 0x1D528623,
            0xF12F560E, 0xF5EE4BB9, 0xF8AD6D60, 0xFC6C70D7,
            0xE22B20D2, 0xE6EA3D65, 0xEBA91BBC, 0xEF68060B,
            0xD727BBB6, 0xD3E6A601, 0xDEA580D8, 0xDA649D6F,
            0xC423CD6A, 0xC0E2D0DD, 0xCDA1F604, 0xC960EBB3,
            0xBD3E8D7E, 0xB9FF90C9, 0xB4BCB610, 0xB07DABA7,
            0xAE3AFBA2, 0xAAFBE615, 0xA7B8C0CC, 0xA379DD7B,
            0x9B3660C6, 0x9FF77D71, 0x92B45BA8, 0x9675461F,
            0x8832161A, 0x8CF30BAD, 0x81B02D74, 0x857130C3,
            0x5D8A9099, 0x594B8D2E, 0x5408ABF7, 0x50C9B640,
            0x4E8EE645, 0x4A4FFBF2, 0x470CDD2B, 0x43CDC09C,
            0x7B827D21, 0x7F436096, 0x7200464F, 0x76C15BF8,
            0x68860BFD, 0x6C47164A, 0x61043093, 0x65C52D24,
            0x119B4BE9, 0x155A565E, 0x18197087, 0x1CD86D30,
            0x029F3D35, 0x065E2082, 0x0B1D065B, 0x0FDC1BEC,
            0x3793A651, 0x3352BBE6, 0x3E119D3F, 0x3AD08088,
            0x2497D08D, 0x2056CD3A, 0x2D15EBE3, 0x29D4F654,
            0xC5A92679, 0xC1683BCE, 0xCC2B1D17, 0xC8EA00A0,
            0xD6AD50A5, 0xD26C4D12, 0xDF2F6BCB, 0xDBEE767C,
            0xE3A1CBC1, 0xE760D676, 0xEA23F0AF, 0xEEE2ED18,
            0xF0A5BD1D, 0xF464A0AA, 0xF9278673, 0xFDE69BC4,
            0x89B8FD09, 0x8D79E0BE, 0x803AC667, 0x84FBDBD0,
            0x9ABC8BD5, 0x9E7D9662, 0x933EB0BB, 0x97FFAD0C,
            0xAFB010B1, 0xAB710D06, 0xA6322BDF, 0xA2F33668,
            0xBCB4666D, 0xB8757BDA, 0xB5365D03, 0xB1F740B4
        };

        internal static int CalculateCrc(byte[] buf, int offs, int size) {     
            uint crc = 0;

            while (--size >= 0) {
                crc = crc32table[((crc >> 24) ^ buf[offs++]) & 255] ^ (crc << 8);
            }
            return (int)crc;
        }
    }
}
