//-< TtreePage.cs >--------------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// C# Program Interface                                              *  /  \  *
//                          Created:     14-Feb-2004  K.A. Knizhnik  * / [] \ *
//                          Last update: 14-Feb-2004  K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Implementation of T-Tree page used by sorted collection
//-------------------------------------------------------------------*--------*
namespace Goods
{
    using System;
    using System.Collections;

    [Metaobject(typeof(OptimisticMetaobject))]
    public class TtreePage : Persistent  
    { 
        const int maxItems = 1000;
        const int minItems = maxItems - 2; // minimal number of items in internal node

        public TtreePage    left;
        public TtreePage    right;
        public int          balance;
        public int          nItems;
        public Persistent[] item;

        public TtreePage(Metaobject mop) : base(mop) {}

        internal TtreePage(Persistent mbr) 
        { 
            nItems = 1;
            item = new Persistent[maxItems];
            item[0] = mbr;
        }

        Persistent loadItem(int i) 
        { 
            return item[i];
        }

        [Constant]
        internal bool find(PersistentComparator comparator, Object minValue, Object maxValue, ArrayList selection)
        { 
            int l, r, m, n;
            n = nItems;
            if (minValue != null) 
            { 
                if (comparator.CompareMemberWithKey(loadItem(0), minValue) < 0) 
                {	    
                    if (comparator.CompareMemberWithKey(loadItem(n-1), minValue) < 0) 
                    { 
                        if (right != null) 
                        { 
                            return right.find(comparator, minValue, maxValue, selection); 
                        } 
                        return true;
                    }
                    for (l = 0, r = n; l < r;) 
                    { 
                        m = (l + r) >> 1;
                        if (comparator.CompareMemberWithKey(loadItem(m), minValue) < 0) 
                        {
                            l = m+1;
                        } 
                        else 
                        { 
                            r = m;
                        }
                    }
                    while (r < n) 
                    { 
                        if (maxValue != null
                            && comparator.CompareMemberWithKey(loadItem(r), maxValue) > 0)
                        { 
                            return false;
                        }
                        selection.Add(loadItem(r));
                        r += 1;
                    }
                    if (right != null) 
                    { 
                        return right.find(comparator, minValue, maxValue, selection); 
                    } 
                    return true;	
                }
            }	
            if (left != null) 
            { 
                if (!left.find(comparator, minValue, maxValue, selection)) 
                { 
                    return false;
                }
            }
            for (l = 0; l < n; l++) 
            { 
                if (maxValue != null && comparator.CompareMemberWithKey(loadItem(l), maxValue) > 0) 
                {
                    return false;
                }
                selection.Add(loadItem(l));
            }
            if (right != null) 
            { 
                return right.find(comparator, minValue, maxValue, selection);
            }         
            return true;
        }
    
        [Constant]
        internal bool contains(PersistentComparator comparator, Persistent mbr)
        { 
            int l, r, m, n;
            n = nItems;
            if (comparator.CompareMembers(loadItem(0), mbr) < 0) 
            {	    
                if (comparator.CompareMembers(loadItem(n-1), mbr) < 0) 
                { 
                    if (right != null) 
                    { 
                        return right.contains(comparator, mbr); 
                    } 
                    return false;
                }
                for (l = 0, r = n; l < r;) 
                { 
                    m = (l + r) >> 1;
                    if (comparator.CompareMembers(loadItem(m), mbr) < 0) 
                    {
                        l = m+1;
                    } 
                    else 
                    { 
                        r = m;
                    }
                }
                while (r < n) 
                { 
                    if (mbr == loadItem(r)) 
                    { 
                        return true;
                    }
                    if (comparator.CompareMembers(item[r], mbr) > 0) 
                    { 
                        return false;
                    }
                    r += 1;
                }
                if (right != null) 
                { 
                    return right.contains(comparator, mbr); 
                } 
                return false;	
            }
            if (left != null) 
            { 
                if (left.contains(comparator, mbr)) 
                { 
                    return true;
                }
            }
            for (l = 0; l < n; l++) 
            { 
                if (mbr == loadItem(l)) 
                { 
                    return true;
                }
                if (comparator.CompareMembers(item[l], mbr) > 0) 
                {
                    return false;
                }
            }
            if (right != null) 
            { 
                return right.contains(comparator, mbr);
            }         
            return false;
        }

    
        internal const int OK         = 0;
        internal const int NOT_UNIQUE = 1;
        internal const int NOT_FOUND  = 2;
        internal const int OVERFLOW   = 3;
        internal const int UNDERFLOW  = 4;

        internal int insert(PersistentComparator comparator, Persistent mbr, bool unique, ref TtreePage pgRef) 
        { 
            int n = nItems;
            TtreePage pg;
            int diff = comparator.CompareMembers(mbr, loadItem(0));
            if (diff <= 0) 
            { 
                if (unique && diff == 0) 
                { 
                    return NOT_UNIQUE;
                }
                if ((left == null || diff == 0) && n != maxItems) 
                { 
                    Modify();
                    for (int i = n; i > 0; i--) item[i] = item[i-1];
                    item[0] = mbr;
                    nItems += 1;
                    return OK;
                } 
                if (left == null) 
                { 
                    Modify();
                    left = new TtreePage(mbr);
                } 
                else 
                {
                    pg = pgRef;
                    pgRef = left;
                    int result = left.insert(comparator, mbr, unique, ref pgRef);
                    if (result == NOT_UNIQUE) 
                    { 
                        return NOT_UNIQUE;
                    }
                    Modify();
                    left = pgRef;
                    pgRef = pg;
                    if (result == OK) return OK;
                }
                if (balance > 0) 
                { 
                    balance = 0;
                    return OK;
                } 
                else if (balance == 0) 
                { 
                    balance = -1;
                    return OVERFLOW;
                } 
                else 
                { 
                    TtreePage lp = this.left;
                    lp.Modify();
                    if (lp.balance < 0) 
                    { // single LL turn
                        this.left = lp.right;
                        lp.right = this;
                        balance = 0;
                        lp.balance = 0;
                        pgRef = lp;
                    } 
                    else 
                    { // double LR turn
                        TtreePage rp = lp.right;
                        rp.Modify();
                        lp.right = rp.left;
                        rp.left = lp;
                        this.left = rp.right;
                        rp.right = this;
                        balance = (rp.balance < 0) ? 1 : 0;
                        lp.balance = (rp.balance > 0) ? -1 : 0;
                        rp.balance = 0;
                        pgRef = rp;
                    }
                    return OK;
                }
            } 
            diff = comparator.CompareMembers(mbr, loadItem(n-1));
            if (diff >= 0) 
            { 
                if (unique && diff == 0) 
                { 
                    return NOT_UNIQUE;
                }
                if ((right == null || diff == 0) && n != maxItems) 
                { 
                    Modify();
                    item[n] = mbr;
                    nItems += 1;
                    return OK;
                }
                if (right == null) 
                { 
                    Modify();
                    right = new TtreePage(mbr);
                } 
                else 
                { 
                    pg = pgRef;
                    pgRef = right;
                    int result = right.insert(comparator, mbr, unique, ref pgRef);
                    if (result == NOT_UNIQUE) 
                    { 
                        return NOT_UNIQUE;
                    }
                    Modify();
                    right = pgRef;
                    pgRef = pg;
                    if (result == OK) return OK;
                }
                if (balance < 0) 
                { 
                    balance = 0;
                    return OK;
                } 
                else if (balance == 0) 
                { 
                    balance = 1;
                    return OVERFLOW;
                } 
                else 
                { 
                    TtreePage rp = this.right;
                    rp.Modify();
                    if (rp.balance > 0) 
                    { // single RR turn
                        this.right = rp.left;
                        rp.left = this;
                        balance = 0;
                        rp.balance = 0;
                        pgRef = rp;
                    } 
                    else 
                    { // double RL turn
                        TtreePage lp = rp.left;
                        lp.Modify();
                        rp.left = lp.right;
                        lp.right = rp;
                        this.right = lp.left;
                        lp.left = this;
                        balance = (lp.balance > 0) ? -1 : 0;
                        rp.balance = (lp.balance < 0) ? 1 : 0;
                        lp.balance = 0;
                        pgRef = lp;
                    }
                    return OK;
                }
            }
            int l = 1, r = n-1;
            while (l < r)  
            {
                int i = (l+r) >> 1;
                diff = comparator.CompareMembers(mbr, loadItem(i));
                if (diff > 0) 
                { 
                    l = i + 1;
                } 
                else 
                { 
                    r = i;
                    if (diff == 0) 
                    { 
                        if (unique) 
                        { 
                            return NOT_UNIQUE;
                        }
                        break;
                    }
                }
            }
            // Insert before item[r]
            Modify();
            if (n != maxItems) 
            {
                for (int i = n; i > r; i--) item[i] = item[i-1]; 
                item[r] = mbr;
                nItems += 1;
                return OK;
            } 
            else 
            { 
                Persistent reinsertItem;
                if (balance >= 0) 
                { 
                    reinsertItem = loadItem(0);
                    for (int i = 1; i < r; i++) item[i-1] = item[i]; 
                    item[r-1] = mbr;
                } 
                else 
                { 
                    reinsertItem = loadItem(n-1);
                    for (int i = n-1; i > r; i--) item[i] = item[i-1]; 
                    item[r] = mbr;
                }
                return insert(comparator, reinsertItem, unique, ref pgRef);
            }
        }
       
        internal int balanceLeftBranch(ref TtreePage pgRef) 
        {
            if (balance < 0) 
            { 
                balance = 0;
                return UNDERFLOW;
            } 
            else if (balance == 0) 
            { 
                balance = 1;
                return OK;
            } 
            else 
            { 
                TtreePage rp = this.right;
                rp.Modify();
                if (rp.balance >= 0) 
                { // single RR turn
                    this.right = rp.left;
                    rp.left = this;
                    if (rp.balance == 0) 
                    { 
                        this.balance = 1;
                        rp.balance = -1;
                        pgRef = rp;
                        return OK;
                    } 
                    else 
                    { 
                        balance = 0;
                        rp.balance = 0;
                        pgRef = rp;
                        return UNDERFLOW;
                    }
                } 
                else 
                { // double RL turn
                    TtreePage lp = rp.left;
                    lp.Modify();
                    rp.left = lp.right;
                    lp.right = rp;
                    this.right = lp.left;
                    lp.left = this;
                    balance = lp.balance > 0 ? -1 : 0;
                    rp.balance = lp.balance < 0 ? 1 : 0;
                    lp.balance = 0;
                    pgRef = lp;
                    return UNDERFLOW;
                }
            }
        }

        internal int balanceRightBranch(ref TtreePage pgRef) 
        {
            if (balance > 0) 
            { 
                balance = 0;
                return UNDERFLOW;
            } 
            else if (balance == 0) 
            { 
                balance = -1;
                return OK;
            } 
            else 
            { 
                TtreePage lp = this.left;
                lp.Modify();
                if (lp.balance <= 0) 
                { // single LL turn
                    this.left = lp.right;
                    lp.right = this;
                    if (lp.balance == 0) 
                    { 
                        balance = -1;
                        lp.balance = 1;
                        pgRef = lp;
                        return OK;
                    } 
                    else 
                    { 
                        balance = 0;
                        lp.balance = 0;
                        pgRef = lp;
                        return UNDERFLOW;
                    }
                } 
                else 
                { // double LR turn
                    TtreePage rp = lp.right;
                    rp.Modify();
                    lp.right = rp.left;
                    rp.left = lp;
                    this.left = rp.right;
                    rp.right = this;
                    balance = rp.balance < 0 ? 1 : 0;
                    lp.balance = rp.balance > 0 ? -1 : 0;
                    rp.balance = 0;
                    pgRef = rp;
                    return UNDERFLOW;
                }
            }
        }
    
        internal int remove(PersistentComparator comparator, Persistent mbr, ref TtreePage pgRef)
        {
            TtreePage pg;
            int n = nItems;
            int diff = comparator.CompareMembers(mbr, loadItem(0));
            if (diff <= 0) 
            { 
                if (left != null) 
                { 
                    Modify();
                    pg = pgRef;
                    pgRef = left;
                    int h = left.remove(comparator, mbr, ref pgRef);
                    left = pgRef;
                    pgRef = pg;
                    if (h == UNDERFLOW) 
                    { 
                        return balanceLeftBranch(ref pgRef);
                    } 
                    else if (h == OK) 
                    { 
                        return OK;
                    }
                }
            }
            diff = comparator.CompareMembers(mbr, loadItem(n-1));
            if (diff <= 0) 
            {	    
                for (int i = 0; i < n; i++) 
                { 
                    if (loadItem(i) == mbr) 
                    { 
                        if (n == 1) 
                        { 
                            if (right == null) 
                            { 
                                //Deallocate();
                                pgRef = left;
                                return UNDERFLOW;
                            } 
                            else if (left == null) 
                            { 
                                //Deallocate();
                                pgRef = right;
                                return UNDERFLOW;
                            } 
                        }
                        Modify();
                        if (n <= minItems) 
                        { 
                            if (left != null && balance <= 0) 
                            {  
                                TtreePage prev = left;
                                while (prev.right != null) 
                                {                                 
                                    prev = prev.right;
                                }
                                while (--i >= 0) 
                                { 
                                    item[i+1] = item[i];
                                }
                                item[0] = prev.item[prev.nItems-1];
                                pg = pgRef;
                                pgRef = left;
                                int h = left.remove(comparator, loadItem(0), ref pgRef);
                                left = pgRef;
                                pgRef = pg;
                                if (h == UNDERFLOW) 
                                {
                                    h = balanceLeftBranch(ref pgRef);
                                }
                                return h;
                            } 
                            else if (right != null) 
                            { 
                                TtreePage next = right;
                                while (next.left != null) 
                                { 
                                    next = next.left;
                                }
                                while (++i < n) 
                                { 
                                    item[i-1] = item[i];
                                }
                                item[n-1] = next.item[0];
                                pg = pgRef;
                                pgRef = right;
                                int h = right.remove(comparator, loadItem(n-1), ref pgRef);
                                right = pgRef;
                                pgRef = pg;
                                if (h == UNDERFLOW) 
                                {
                                    h = balanceRightBranch(ref pgRef);
                                }
                                return h;
                            }
                        }
                        while (++i < n) 
                        { 
                            item[i-1] = item[i];
                        }
                        nItems -= 1;
                        return OK;
                    }
                }
            }
            if (right != null) 
            { 
                Modify();
                pg = pgRef;
                pgRef = right;
                int h = right.remove(comparator, mbr, ref pgRef);
                right = pgRef;
                pgRef = pg;
                if (h == UNDERFLOW) 
                { 
                    return balanceRightBranch(ref pgRef);
                }
                else 
                { 
                    return h;
                }
            }
            return NOT_FOUND;
        }


        [Constant]
        internal int toArray(Persistent[] arr, int index) 
        { 
            if (left != null) 
            { 
                index = left.toArray(arr, index);
            }
            for (int i = 0, n = nItems; i < n; i++) 
            { 
                arr[index++] = loadItem(i);
            }
            if (right != null) 
            { 
                index = right.toArray(arr, index);
            }
            return index;
        }

        internal void prune() 
        { 
            if (left != null) 
            { 
                left.prune();
            }
            if (right != null) 
            { 
                right.prune();
            }
            // Deallocate();
        }
    }
}
