#include "goods.h"

USE_GOODS_NAMESPACE

typedef nat8 skey_t;

#define N_ITEMS 128

class page : public object { 
  public: 
    int4 m;            // Index of first used item

#define VARYING_PAGE
#ifdef VARYING_PAGE
    struct item { 
        ref<page> p; // for leaf page - pointer to 'set_member' object
                       // for all other pages - pointer to child page
        skey_t    key; 

        field_descriptor& describe_components() { 
            return FIELD(p), FIELD(key);
        } 
    } e[1];

    static page* create(int len) { 
        return new (self_class, len) page(len);
    }

    page(size_t varying_length) : object(self_class, varying_length) {
        m = varying_length;
    }

#else
    struct item { 
        ref<page> p; // for leaf page - pointer to 'set_member' object
                       // for all other pages - pointer to child page
        skey_t    key; 

        field_descriptor& describe_components() { 
            return FIELD(p), FIELD(key);
        } 
    } e[N_ITEMS];

    static page* create(int) { 
        return new page();
    }

    page() : object(self_class) {
        m = N_ITEMS;
    }

#endif

    void init() { 
        for (int i = 0; i < m; i++) { 
            e[i].key = i;
            e[i].p = this;
        }
    }

    void dump() const; 
    friend field_descriptor& describe_field(item& s);

    METACLASS_DECLARATIONS(page, object);
};

REGISTER(page, object, pessimistic_repeatable_read_scheme);

void page::dump() const { 
    if (is_abstract_root()) { 
        page* p = (page*)modify(this)->become(create(N_ITEMS));
       modify(p)->init();
    } else { 
        assert(m == N_ITEMS);
        for (int i = 0; i < m; i++) { 
            assert(e[i].key == i);
            assert(e[i].p == this);
        }
    }
}

inline field_descriptor& describe_field(page::item& s) { 
    return s.describe_components();
}

#ifdef VARYING_PAGE
field_descriptor& page::describe_components()
{
    return FIELD(m), VARYING(e); 
}
#else
field_descriptor& page::describe_components()
{
    return FIELD(m), ARRAY(e); 
}
#endif

int main() 
{
    task::initialize(task::huge_stack);
    ref<page> root;
    database db;

    if (db.open("testeval.cfg")) { 
        db.get_root(root);
        root->dump();
        db.close();
        printf("Test passed\n");
        return 0;
    }
    return 1;
}
    

