//-< RawBinaryFactory.java >-----------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:     26-Sep-2002  K.A. Knizhnik  * / [] \ *
//                          Last update: 26-Oct-2002  K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Interace of factory for packing/unpacking raw binary objects
//-------------------------------------------------------------------*--------*

package goodsjpi;

/**
 * Interace of factory for packing/unpacking raw binary objects
 */
public interface RawBinaryFactory { 
    /**
     * Get size of raw binary object in bytes
     * @param obj raw binary object
     * @return size of packed raw binary object
     */
    public int size(Object obj);
    
    
    /**
     * Pack raw binary object to byte array
     * @param obj raw binary object
     * @param arr array to store packed image of raw binary object
     * @param off offset in the array where object should be packed
     * @return bumer of bytes in packed object image.
     * It should be the same as returned by <code>RawBinary.size</code> method.
     */
    public int pack(Object obj, byte[] arr, int off); 

    /**
     * Unpack raw binary object
     * @param cls class of raw binary object
     * @param arr array containing packed raw binary object
     * @param off offset of packed binary object in the array
     * @param size size of packed binary object
     * @return created raw binary object with unpacked content
     */
    public Object unpack(Class cls, byte[] arr, int off, int size);

    /**
     * Check if specified las supported by thisd factory
     * @param cls raw binary object class
     * @return <code>true</code> if factory is able to pack/unpack instances of these class
     */
    public boolean supports(Class cls);
};
