package org.dcg.gen;

import goodsjpi.Assert;
import goodsjpi.Database;

import org.dcg.datamgr.DBConnector;
import org.dcg.datamgr.databasejobs.DatabaseJob;
import org.dcg.datamgr.goodsindex.RootClass;

import org.dcg.gen.GeneralJob;


/**
 * @author dacosta
 * Each thread in the pool is an instance of this class.<br>
 * This thread-instance is actually 'the thing' that executes the job handed to it.<br>
 * An example would be an employee in a company.<br>
 * Basically this is the 'worker/ labour class' that is available to the Mgr/ the pool to have jobs exectuted.<br>
 */
public abstract class ThreadPoolWorker extends Thread
{

	public boolean _idle; // if true then the thread can process a new job
	private Object _lock; // This object serves as the key to the controlling PoolMgr. It is used to control actions ON the PoolMgr
	private Object _data; // This is the Object that carries the date with which the work needs to be done

	//private Worker _worker; // _worker is the actual 'thing' on which the thread is going to work.
	private DBConnector _databaseConnector = null;
	private Database _database; // This is the representation of the database in the WorkerThread
	private RootClass _root; // In case of an OODBMS this is the root object to be used to gain access to the rest of the structure

	/** Override the following attributes in the PoolMgr.<br>*/
	protected static int __threadCounter;
	protected static boolean __debug = false;
	protected static int __maxIdleTime;
	protected static int __minThreads;
	protected static int __maxThreads; 

	/**
	 * Normal constructor for a NON-database PoolMgr.<br> Pass in the PoolMgr instance for synchronization.<br>
	 * 
	 * @param _lock represents the PoolMgr
	 */
	public ThreadPoolWorker(ThreadGroup tg, Object lock, boolean isDaemon)
	{
		super(tg, "Thread-" + String.valueOf(__threadCounter++));
		_lock = lock;
		_idle = true;

		if (isDaemon) {
			this.setDaemon(true);
		}

		//setParentAttributes();
		//this.start();
	}

	/**
	 * Constructor for a database PoolMgr.<br> Pass in the PoolMgr instance for synchronization.<br>
	 * 
	 * @param _lock represents the PoolMgr
	 */
	public ThreadPoolWorker(ThreadGroup tg, Object lock, boolean isDaemon, DBConnector dbConnector)
	{
		super(tg, "Thread-" + String.valueOf(__threadCounter++));
		_lock = lock;
		_idle = true;

		if (isDaemon) {
			this.setDaemon(true);
		}

		//setParentAttributes();
		//this.start();
		_databaseConnector = dbConnector;

		if (__debug) {
			System.out.println(this.toString() + " constructor of the PoolThreadWorker");
		}

		// Open the GOODS database
		_database = new Database();

		boolean success = _database.open(dbConnector.getDatabaseConfigFile());
		Assert.that(success);
		_root = (RootClass)_database.getRoot();

		if (_root == null) {
			_root = new RootClass();
			_database.setRoot(_root);
		}
    }
	

	protected abstract void setParentAttributes();

	/**
	 * int _threadCounter = 0; boolean _debug =false; int _maxIdleTime=0; int _minThreads=0;
	 */
	/**
	 * Wakes the thread and does some work
	 * 
	 * @param data Data to send to the Worker
	 * @return void
	 */
	public synchronized void wake(Object data)
	{
		_data = data;
		if (__debug) { System.out.println(data.toString()); }
		notify();
	}

	// This is where all the action is...

	/**
	 * DOCUMENT ME!
	 */
	public void run()
	{

		Runnable job = null;
		System.err.println("Entered the run/ Start of the thread: " + this.getName());
		System.err.println("The number of jobs pending is: " + getNrOfPendingJobs());

		while (true) { // While I
			while (true) { // While II

				synchronized (_lock) {

					// Keep processing jobs until none availble
					if (getNrOfPendingJobs() == 0) {

						if (__debug) {
							System.out.println("Push idle " + this.getName() + " on queue with size: " + getAvailableThreads());
						}
						this._idle = true;
						break;
						// Jump out of the while loop because there are no more jobs in  the queue
					}
					// Remove the job from the pending list.
					if (_databaseConnector == null) {
						job = (GeneralJob)getPendingJob();
					} else {
						job = (DatabaseJob)getPendingJob();
					}
				}

				// Run the job (later insert the data to be worked on)
				this._idle = false; // Put a lock on the Mgr so that no other thread can gain access at the same time to go into the job queue

				long start = System.currentTimeMillis();

				if (_databaseConnector == null) {
					((GeneralJob)job).run();
				} else {
					((DatabaseJob)job).run(_database, _root);
				}

				//if (__debug) System.out.println(">>> Finished job "+job.toString()+" on: "+ this);
				System.out.println("ThreadPoolMgrWorker.run()=> Finished job in: " + (System.currentTimeMillis() - start) + " mSecs");

				//System.out.println("Abs TPWorker.run() >>> Just came out of "+Thread.currentThread().getName()+ " with "+ ((DatabaseJob)job).toString());
				job = null;
			} // While II

			try {

				synchronized (this) {

					// If no idle time specified, wait till notified.
					if (__maxIdleTime == -1) {
						pushIntoAvailableThreads(this);

						if (__debug) {
							System.out.println("Push " + this.getName() + " on ALWAYS idle queue NEW size: " + getAvailableThreads());
						}

						wait();
					} else {

						if (!currentThreadInAvailableThreads(this)) {
							pushIntoAvailableThreads(this);
						}

						if (__debug) {
							System.out.println("Pushed " + this.getName() + " on MAXIDLE queue NEW size: " + getAvailableThreads());
						}

						wait(__maxIdleTime);
					}
				}
			} catch (InterruptedException e) {

				// Clean up if interrupted
				synchronized (_lock) {

					if (__debug) {
						System.out.println("Interrupted so clean-up: " + this.getName());
					}

					if (_database != null) {
						_database.close();
					}

					removeCurrentThreadFromQueue(this);
				}

				return;
			}

			// Just been notified or the wait timed out
			synchronized (_lock) {

				// If there are no jobs, that means we "idled" out.
				if (getNrOfPendingJobs() == 0) {

					if ((__minThreads != -1) && (getAvailableThreads() > __minThreads)) {

						// For complete runout of prg do >= for ALWAYS pending threads just do >
						if (__debug) {
							System.out.println(this.getName() + " timed out so remove=> pool size: " + getAvailableThreads());
						}

						if (_database != null) {
							_database.close();
						}

						removeCurrentThreadFromQueue(this);

						return;
					}
				}
			}
		} // While( true ) I
	}

	protected abstract int getNrOfPendingJobs(); // _pendingJobs.size()

	protected abstract Object getPendingJob(); // _pendingJobs.pop()

	protected abstract int getAvailableThreads(); // _availableThreads.size()

	protected abstract void pushIntoAvailableThreads(ThreadPoolWorker pw); // _availableThreads.push(this);

	protected abstract boolean currentThreadInAvailableThreads(ThreadPoolWorker pw); // _availableThreads.contains(this);

	protected abstract void removeCurrentThreadFromQueue(ThreadPoolWorker pw); // removeMe(ths)
}
